//===========================================================================
// Copyright  1998 Thin Air Enterprises and Robert Dunn
//===========================================================================
//---------------------------------------------------------------------------
// TRichEdit20 implementation
//---------------------------------------------------------------------------
//#include <vcl\vcl.h>
#include <vcl.h>
#pragma hdrstop

#include "richedit20.h"

#ifdef MAX_TAB_STOPS
#undef MAX_TAB_STOPS
#define MAX_TAB_STOPS 32
#endif
//---------------------------------------------------------------------------
// static member functions
//
bool TRichEdit20::IsRTF(AnsiString filepath)
{
	static char idBytes[5] = "{\\rtf";
	char buf[sizeof(idBytes)];
	int fileHandle;
	int bytesRead;
	bool result = false;

	try {
		fileHandle = FileOpen(filepath, fmOpenRead | fmShareDenyWrite);
		bytesRead = FileRead(fileHandle, buf, sizeof(buf));
		FileClose(fileHandle);
		if (bytesRead == sizeof(buf)) result = !memcmp(idBytes, buf, sizeof(buf));
		}
	catch (...) {
		if (fileHandle) FileClose(fileHandle);
		}
	return result;
}
//---------------------------------------------------------------------------
// check if a file is likely to be a binary file -- arbitrarily considered
// binary if has a null byte in first 8192 bytes of file
//
bool TRichEdit20::IsLikelyBinaryFile(AnsiString filepath)
{
	char buf[8192];
	::memset(buf, 'z', sizeof(buf));

	int fileHandle;
	bool result = false;

	try {
		fileHandle = FileOpen(filepath, fmOpenRead | fmShareDenyWrite);
		FileRead(fileHandle, buf, sizeof(buf));
		FileClose(fileHandle);
		result = memchr(buf, '\0', sizeof(buf));
		}
	catch (...) {
		if (fileHandle) FileClose(fileHandle);
		}
	return result;
}
//---------------------------------------------------------------------------
__fastcall TRichEdit20::TRichEdit20(TComponent* Owner)
#ifdef USERICHEDITOLE
	: TRichEdit(Owner), FInsertMode(true), FOnInsertChange(0),
		FRichEditOle(0), FAutoUrlDetect(false), FSelAttributes2(0),
		FDefAttributes2(0)
#else
	: TRichEdit(Owner), FInsertMode(true), FOnInsertChange(0),
		FAutoUrlDetect(false), FSelAttributes2(0), FDefAttributes2(0)
#endif
{
	// allocate text attribute members
	FSelAttributes2 = new TTextAttributes2(this, atSelected);
	FDefAttributes2 = new TTextAttributes2(this, atDefaultText);
	FParaAttributes2 = new TParaAttributes2(this);

	// set some defaults
	FTabStopsCount = 0;
	for (int i = 0; i < MAX_TAB_STOPS; i++) FTabStops[i] = 0;
}
//---------------------------------------------------------------------------
__fastcall TRichEdit20::~TRichEdit20()
{
	if (FSelAttributes2) delete FSelAttributes2;
	if (FDefAttributes2) delete FDefAttributes2;
	if (FParaAttributes2) delete FParaAttributes2;
}
//---------------------------------------------------------------------------
TRect TRichEdit20::TwipsToTargetRect(HDC hdc, TRect rect)
{
	int logx = ::GetDeviceCaps(hdc, LOGPIXELSX);
	int logy = ::GetDeviceCaps(hdc, LOGPIXELSY);

	TRect r;
	r.Left = ::MulDiv(rect.Left, logx, 1440);
	r.Top = ::MulDiv(rect.Top, logy, 1440);
	r.Right = ::MulDiv(rect.Right, logx, 1440);
	r.Bottom = ::MulDiv(rect.Bottom, logy, 1440);

	return r;
}
//---------------------------------------------------------------------------
bool __fastcall TRichEdit20::GetCanUndo(void)
{
	return ::SendMessage(Handle, EM_CANUNDO, 0, 0);
}
//---------------------------------------------------------------------------
bool __fastcall TRichEdit20::GetCanRedo(void)
{
	return ::SendMessage(Handle, EM_CANREDO, 0, 0);
}
//---------------------------------------------------------------------------
bool TRichEdit20::Undo(void)
{
	// returns zero on failure
	return ::SendMessage(Handle, EM_UNDO, 0, 0);
}
//---------------------------------------------------------------------------
bool TRichEdit20::Redo(void)
{
	// returns zero on failure
	return ::SendMessage(Handle, EM_REDO, 0, 0);
}
//---------------------------------------------------------------------------
void __fastcall TRichEdit20::SetUndoLimit(int limit)
{
	// returns number actually set -- may be smaller than request
	FUndoLimit = ::SendMessage(Handle, EM_SETUNDOLIMIT, limit, 0);
}
//---------------------------------------------------------------------------
void TRichEdit20::StopGroupTyping(void)
{
	// forces new group for undo -- cannot fail
	::SendMessage(Handle, EM_STOPGROUPTYPING, 0, 0);
}
//---------------------------------------------------------------------------
// build a string of text representing the nature of the next undo/redo
//
static const char *UndoNames[] = {
	"Cannot ", " Typing", " Delete", " Drop Object", " Cut", " Paste"
	};

AnsiString TRichEdit20::BuildUndoName(AnsiString typeName, int id)
{
	if (id >= sizeof(UndoNames) / sizeof(UndoNames[0])) id = 0;
	AnsiString s(UndoNames[id]);
	if (!id) return s + typeName;
	else return typeName + s;
}
//---------------------------------------------------------------------------
// build a string of text representing the nature of the next undo operation
//
AnsiString __fastcall TRichEdit20::GetUndoName(void)
{
	int id = ::SendMessage(Handle, EM_GETUNDONAME, 0, 0);
	if (CanUndo && !id) return "Undo";
	return BuildUndoName("Undo", id);
}
//---------------------------------------------------------------------------
// build a string of text representing the nature of the next redo operation
//
AnsiString __fastcall TRichEdit20::GetRedoName(void)
{
	int id = ::SendMessage(Handle, EM_GETREDONAME, 0, 0);
	if (CanRedo && !id) return "Redo";
	return BuildUndoName("Redo", id);
}
//---------------------------------------------------------------------------
// tell richedit to handle/not handle multiple undos/redos
//
void __fastcall TRichEdit20::SetMultiUndo(bool multiLevel)
{
	// success == 0, failure != 0
	FSupportsMultiUndo = ::SendMessage(Handle, EM_SETTEXTMODE,
		(WPARAM) TM_MULTILEVELUNDO, 0);
	if (FSupportsMultiUndo)
		FMultiUndo = ::SendMessage(Handle, EM_SETTEXTMODE,
			(WPARAM) multiLevel ? TM_MULTILEVELUNDO : TM_SINGLELEVELUNDO, 0);
}
//---------------------------------------------------------------------------
// do stuff when creating/recreating window
//
void __fastcall TRichEdit20::CreateWnd(void)
{
	TRichEdit::CreateWnd();

	// set up ole callback
#ifdef USERICHEDITOLE
	FRichEditOle = new TIRichEditOle(this);
#endif

	MultiUndo = true;							// initialize
	if (FSupportsMultiUndo) UndoLimit = 100;	//???
	else FUndoLimit = 0;

	// set tabstops, if needed
	if (FTabStopsCount) {
		TTabStops tabs;
		GetTabStops(tabs);
		SetTabStops(FTabStopsCount, tabs);
		}

	// set insert state to match property
	if (!FInsertMode) {
		FInsertMode = true;	// this is the starting state for RichEdits
		ToggleInsertMode(); // this will toggle shadow variable (FInsertMode)
		}

	// set auto url detection to match property state
	SetAutoUrlDetect(FAutoUrlDetect);
}
//---------------------------------------------------------------------------
// the following ensures that the new version of richedit (v2.0) is loaded;
// TRichEdit unloads the DLL that *IT* loads in the WM_NCDESTROY handler (which,
// disappointingly enough, is private rather than protected; damned if FLibHandle
// isn't also, or we could simply unload it later) but we do it in DestroyWnd()...
// (note:  do not call TRichEdit::CreateParams -- it will attempt to
// load the older version of the rich edit DLL -- not a Good Thing -- too bad
// if anything important goes on there)
//
void __fastcall TRichEdit20::CreateParams(Controls::TCreateParams &Params)
{
	const char RichEditModuleName[] = "RICHED20.DLL";
	long int OldError;

	OldError = SetErrorMode(SEM_NOOPENFILEERRORBOX);
	FLibHandle = (int) LoadLibrary(RichEditModuleName);
	if (FLibHandle < HINSTANCE_ERROR) FLibHandle = 0;
	SetErrorMode(OldError);
//	TRichEdit::CreateParams(Params);
	TCustomMemo::CreateParams(Params);
//	CreateSubClass(Params, "RICHEDIT");
	CreateSubClass(Params, RICHEDIT_CLASS);
	Params.Style = Params.Style |
		(HideScrollBars ? 0 : ES_DISABLENOSCROLL) |
		(HideSelection ? 0 : ES_NOHIDESEL);
}
//---------------------------------------------------------------------------
// free the library on destroy (actually, it will not unload until last
// TRichEdit20 instance is destroyed)
//
void __fastcall TRichEdit20::DestroyWnd(void)
{
#ifdef USERICHEDITOLE
	if (FRichEditOle) {
		delete FRichEditOle;
		FRichEditOle = 0;
		}
#endif

	TRichEdit::DestroyWnd();

	if (FLibHandle != 0) FreeLibrary((void*) FLibHandle);
}
//---------------------------------------------------------------------------
CHARRANGE __fastcall TRichEdit20::GetSelRange(void)
{
	CHARRANGE chrg;
	::SendMessage(Handle, EM_EXGETSEL, 0, (LPARAM) &chrg);
	return chrg;
}
//---------------------------------------------------------------------------
// trap link messages and execute link
//
void __fastcall TRichEdit20::WndProc(Messages::TMessage &Message)
{
	// is this a control notification message?
	if (Message.Msg == CN_NOTIFY) {
		// is this an EN_LINK notification?
		TWMNotify& msg = *((TWMNotify*) (void*) &Message);
		if (msg.NMHdr->code == EN_LINK) {
			// is this a left button down notification?
			TENLink& enLink = *((TENLink*) msg.NMHdr);
			if (enLink.msg == WM_LBUTTONDOWN) {
				// select the text
				::SendMessage(Handle, EM_EXSETSEL, 0, (LPARAM) &enLink.chrg);
				// allocate a buffer for the text
				char* buf = new char[enLink.chrg.cpMax - enLink.chrg.cpMin + 1];
				// get the text
				::SendMessage(Handle, EM_GETSELTEXT, 0, (LPARAM) buf);
				// execute the link
				::ShellExecute(Handle, "open", buf, "", "", SW_SHOWNORMAL);
				// deallocate the buffer
				delete buf;
				// return non-zero to tell the RE that we handled this message
				Message.Result = !0;
				return;
				}
			}
		}
		
	TRichEdit::WndProc(Message);
}
//---------------------------------------------------------------------------
void TRichEdit20::ScrollCaret(void)
{
	::SendMessage(Handle, EM_SCROLLCARET, 0, 0L);
}
//---------------------------------------------------------------------------
bool TRichEdit20::FindTextEx(AnsiString text, int startPos,
	TFindOptions findOptions, int& foundPos, int& foundLength)
{
	FINDTEXTEX ft;
	ft.chrg.cpMin = startPos;		// start pos for search
	ft.chrg.cpMax = -1;				// search all from start pos
	ft.lpstrText = text.c_str();	// better not modify!!!
	ft.chrgText.cpMin = -1;			// return start
	ft.chrgText.cpMax = -1;			// return end

	WPARAM flags = 0;
	if (findOptions.Contains(frDown)) flags |= FR_DOWN;				// and this is documented WHERE???
	if (findOptions.Contains(frMatchCase)) flags |= FT_MATCHCASE;	// apparently could use FR_MATCHCASE
	if (findOptions.Contains(frWholeWord)) flags |= FT_WHOLEWORD;	// apparently could use FR_WHOLEWORD

	int pos = ::SendMessage(Handle, EM_FINDTEXTEX, flags, (LPARAM) (FINDTEXTEX FAR *) &ft);

	if (pos != -1) {
		foundPos = ft.chrgText.cpMin;
		foundLength = ft.chrgText.cpMax - foundPos;
		}
	return pos != -1;
}
//---------------------------------------------------------------------------
void __fastcall TRichEdit20::GetExSel(int &SelStart, int &SelStop)
{
	CHARRANGE chrg;
	::SendMessage(Handle, EM_EXGETSEL, 0, (LPARAM) &chrg);
	SelStart = chrg.cpMin;
	SelStop = chrg.cpMax;
}
//---------------------------------------------------------------------------
void __fastcall TRichEdit20::SetExSel(int SelStart, int SelStop)
{
	CHARRANGE chrg;
	chrg.cpMin = SelStart;
	chrg.cpMax = SelStop;
	::SendMessage(Handle, EM_EXSETSEL, 0, (LPARAM) &chrg);
}
//---------------------------------------------------------------------------
// assigning text to the TRichEdit::SelText property replaces the selection
// in a way that the change cannot be undone (maybe uses same call as
// this, but WPARAM = false?).  this version enables the undo.
//
void __fastcall TRichEdit20::ReplaceSelectedText(AnsiString text, bool canUndo)
{
	::SendMessage(Handle, EM_REPLACESEL, (WPARAM) canUndo, (LPARAM) text.c_str());
}
//---------------------------------------------------------------------------
#ifdef USERICHEDITOLE
void __fastcall TRichEdit20::PasteSpecialFromClipboard(unsigned int format)
{
	if (PlainText) format = CF_TEXT;
	::SendMessage(Handle, EM_PASTESPECIAL, (WPARAM) format, 0);
}
//---------------------------------------------------------------------------
bool __fastcall TRichEdit20::PasteSpecial(void)
{
	return FRichEditOle->PasteSpecial();
}
//---------------------------------------------------------------------------
bool __fastcall TRichEdit20::InsertObject(void)
{
	return FRichEditOle->InsertObject();
}
#endif
//---------------------------------------------------------------------------
bool TRichEdit20::CanPasteFormat(unsigned int format)
{
	if (PlainText && format != CF_TEXT) return false;
	return ::SendMessage(Handle, EM_CANPASTE, (WPARAM) format, 0);
}
//---------------------------------------------------------------------------
// is this used/needed???
//
void __fastcall TRichEdit20::RecreateWnd(void)	// not virtual in base!!!
{
	TRichEdit::RecreateWnd();
}
//---------------------------------------------------------------------------
// apparently there is a bug which we must work around... if you drop a large
// file (> 32k?) on a TRichEdit (or, at least, a TRichEdit20), the
// vertical scroll bar is not enabled until the control is manually scrolled
// down a page.  this would not be such a big deal if we trapped the
// LoadFromFile() used, typically, to load the TRichEdit control.  unfortunately,
// that would require deriving from TRichEditStrings and working a lot harder
// than I'm willing to work at the moment, so we'll try to fix it here.  the
// sad result is that the programmer must call this function after loads to
// make sure the vertical scroll bar is visible.  man!  working around
// Microsoft bugs is a lot easier in OWL, but it (OWL) looks like a dead horse.
// anyway, it is an ugly work-around -- it has an ugly name...
//
void TRichEdit20::WorkAroundVerticalScrollBarBug(void)
{
	// scroll down a page, then back
	::SendMessage(Handle, EM_SCROLL, (WPARAM) (INT) SB_PAGEDOWN, 0);
	::SendMessage(Handle, EM_SCROLL, (WPARAM) (INT) SB_PAGEUP, 0);
}
//---------------------------------------------------------------------------
int TRichEdit20::GetTabStops(TTabStops& tabStops)
{
	for (int i = 0; i < MAX_TAB_STOPS; i++) tabStops[i] = FTabStops[i];
	return FTabStopsCount;
}
//---------------------------------------------------------------------------
void TRichEdit20::SetTabStops(int tabCount, TTabStops& tabStops)
{
	PARAFORMAT2 pfmt;
	pfmt.cbSize = sizeof(pfmt);

	pfmt.dwMask = PFM_TABSTOPS;

	for (int i = 0; i < MAX_TAB_STOPS; i++) {
		pfmt.rgxTabs[i] = tabStops[i];
		FTabStops[i] = tabStops[i];
		}

	pfmt.cTabCount = (SHORT) tabCount;
	FTabStopsCount = tabCount;

	if (::IsWindow(WindowHandle))
		::SendMessage(Handle, EM_SETPARAFORMAT, 0, (LPARAM) &pfmt);
}
//---------------------------------------------------------------------------
void TRichEdit20::SetTabStops(int tabSize)
{
	TTabStops tabs;
	for (int i = 0; i < MAX_TAB_STOPS; i++) tabs[i] = (i + 1) * tabSize;

	SetTabStops(MAX_TAB_STOPS, tabs);
}
//---------------------------------------------------------------------------
void TRichEdit20::SetTabStops(int tabChars, TFont& font, bool entireDocument)
{
	HDC hdc = ::GetDC(Handle);
	::SaveDC(hdc);

	::SetMapMode(hdc, MM_TEXT);

	TFont* afont = new TFont();
	afont->Assign(&font);
	::SelectObject(hdc, afont->Handle);

	AnsiString s = AnsiString::StringOfChar('X', tabChars);
	SIZE sz;
	volatile cx = GetDeviceCaps(hdc, LOGPIXELSX);
	int success = ::GetTextExtentPoint32(hdc, s.c_str(), tabChars, &sz);

	int tabSizeTwips = sz.cx;

	::RestoreDC(hdc, -1);
	::ReleaseDC(Handle, hdc);

	delete afont;

	tabSizeTwips = ::MulDiv(tabSizeTwips, 1440, cx);

	// hide and save selection
	CHARRANGE chrg, chrgAll;
	::SendMessage(Handle, EM_EXGETSEL, 0, (LPARAM) &chrg);

	// if entireDocument, set selection to all
	if (entireDocument) {
		chrgAll.cpMin = 0;
		chrgAll.cpMax = -1;
		::SendMessage(Handle, EM_EXSETSEL, 0, (LPARAM) &chrgAll);
		}

	// set tab size
	if (success) SetTabStops(abs(tabSizeTwips));	// no error on failure...

	// restore selection
	::SendMessage(Handle, EM_EXSETSEL, 0, (LPARAM) &chrg);
}
//---------------------------------------------------------------------------
unsigned long int TRichEdit20::GetCurrentLength(void)
{
	// get the length of the text in the control (see treatise above)
	GETTEXTLENGTHEX gtlx = { GTL_PRECISE, CP_ACP };
	return ::SendMessage(Handle, EM_GETTEXTLENGTHEX, (WPARAM) &gtlx, 0);
}
//---------------------------------------------------------------------------
#ifdef USERICHEDITOLE
void TRichEdit20::SetOleHostNames(AnsiString hostApp, AnsiString hostDoc)
{
	FRichEditOle->SetHostNames(hostApp, hostDoc);
}
#endif
//---------------------------------------------------------------------------
// track insert/overstrike state
//
void __fastcall TRichEdit20::KeyDown(Word &Key, Classes::TShiftState Shift)
{
	TRichEdit::KeyDown(Key, Shift);

	TShiftState noShiftKeys;
	if (Key == VK_INSERT && Shift == noShiftKeys) {
		FInsertMode = !FInsertMode;
		if (FOnInsertChange) FOnInsertChange(this);
		}
}
//---------------------------------------------------------------------------
// note: SetInsertMode() does not trigger an OnInsertChange event
//
void __fastcall TRichEdit20::SetInsertMode(bool insertMode)
{
	if (insertMode == FInsertMode) return;
	ToggleInsertMode();
}
//---------------------------------------------------------------------------
// note: ToggleInsertMode() does not trigger an OnInsertChange event --
// the FInsertMode member shadow variable gets flipped in the KeyDown()
// handler
//
void TRichEdit20::ToggleInsertMode(void)
{
	// synthesize an insert keystroke (cannot use keybd_event() because
	// there is no window associated with the API function)
	if (!Handle) return;
	// save and clear the event handler
	TNotifyEvent event = FOnInsertChange;
	FOnInsertChange = 0;
	// the following scarfed from a Micro$oft VB(?) example
	SendMessage(Handle, WM_KEYDOWN, VK_INSERT, 0x00510001);
	SendMessage(Handle, WM_KEYUP, VK_INSERT, 0xC0510001);
	// restore the event handler
	FOnInsertChange = event;
}
//---------------------------------------------------------------------------
unsigned int TRichEdit20::GetLine(void)
{
	CHARRANGE chrg;
	SendMessage(Handle, EM_EXGETSEL, 0, (LPARAM) &chrg);
	return SendMessage(Handle, EM_EXLINEFROMCHAR, 0, (LPARAM) chrg.cpMin);
}
unsigned int TRichEdit20::GetColumn(void)
{
	CHARRANGE chrg;
	SendMessage(Handle, EM_EXGETSEL, 0, (LPARAM) &chrg);
	return chrg.cpMin - SendMessage(Handle, EM_LINEINDEX, (WPARAM) GetLine(), 0);
}
void TRichEdit20::SetLine(unsigned int line)
{
	CHARRANGE chrg;
	chrg.cpMin = chrg.cpMax = SendMessage(Handle, EM_LINEINDEX, (WPARAM) line, 0);
	if (chrg.cpMin == -1)	// past end of text so find last line
		chrg.cpMin = chrg.cpMax = GetCurrentLength();
	SendMessage(Handle, EM_EXSETSEL, 0, (LPARAM) &chrg);
}
void TRichEdit20::SetColumn(unsigned int column)
{
	CHARRANGE chrg;
	unsigned int currLine = GetLine();
	unsigned int ndx = SendMessage(Handle, EM_LINEINDEX, (WPARAM) currLine, 0);
	unsigned int cols = SendMessage(Handle, EM_LINEINDEX, (WPARAM) currLine + 1, 0) -
		ndx - 1;
	if (column > cols) column = cols;
	chrg.cpMin = chrg.cpMax = ndx + column;
	SendMessage(Handle, EM_EXSETSEL, 0, (LPARAM) &chrg);
}
//---------------------------------------------------------------------------
#ifdef USERICHEDITOLE
bool TRichEdit20::CloseActiveObjects(bool savePrompt)
{
	return FRichEditOle->CloseActiveObjects(savePrompt);
}
#endif
//---------------------------------------------------------------------------
void __fastcall TRichEdit20::SetAutoUrlDetect(bool autoDetect)
{
	// tell the RE to auto detect URLs
	::SendMessage(Handle, EM_AUTOURLDETECT, autoDetect, (LPARAM) 0);

	// get the current event notification mask
	unsigned int mask = ::SendMessage(Handle, EM_GETEVENTMASK, 0, 0);

	// set or clear the link notification bit
	if (autoDetect) mask |= ENM_LINK;
	else mask &= ~ENM_LINK;

	// set the new event notification mask
	::SendMessage(Handle, EM_SETEVENTMASK, 0, mask);
}
//---------------------------------------------------------------------------
void __fastcall TRichEdit20::SetSelAttributes2(TTextAttributes2* attr)
{
	SelAttributes2->Assign(attr);
}
//---------------------------------------------------------------------------
void __fastcall TRichEdit20::SetDefAttributes2(TTextAttributes2* attr)
{
	DefAttributes2->Assign(attr);
}
//---------------------------------------------------------------------------
// end of TRichEdit20 implementation
//---------------------------------------------------------------------------
//===========================================================================
// Copyright  1998 Thin Air Enterprises and Robert Dunn
//===========================================================================
