//===========================================================================
// Copyright  1998 Thin Air Enterprises and Robert Dunn
//===========================================================================
//---------------------------------------------------------------------------
#ifndef RichEdit20H
#define RichEdit20H
//---------------------------------------------------------------------------
#ifdef _RICHEDIT_VER
#undef _RICHEDIT_VER
#endif
#define _RICHEDIT_VER	0x0200

#include <vcl\SysUtils.hpp>
#include <vcl\Controls.hpp>
#include <vcl\Classes.hpp>
#include <vcl\Forms.hpp>
#include <vcl\ComCtrls.hpp>
#include <vcl\StdCtrls.hpp>

#include <vector>
#include <values.h>

#include "richedit20format.h"

// comment out the following if you are not using TRichEditOLE and
// remove RichEditOle and RichEditOleCallback from the project.
#define USERICHEDITOLE

#ifdef USERICHEDITOLE
#include "richeditole.h"
#endif
//---------------------------------------------------------------------------
typedef int TTabStops[MAX_TAB_STOPS];

// need to change to derive from TCustomRichEdit and publish only
// those properties desired -- for example, this code will probably
// end up with a different implementation of tabs
//
class TRichEdit20 : public TRichEdit
{
private:
	int FLibHandle;

protected:
	bool FSupportsMultiUndo;
	int FUndoLimit;
	bool FMultiUndo;
	int FTabStopsCount;
	TTabStops FTabStops;
	bool FInsertMode;
	TNotifyEvent FOnInsertChange;
	TNotifyEvent FOnUrlClick;
	bool FAutoUrlDetect;
	TTextAttributes2* FSelAttributes2;
	TTextAttributes2* FDefAttributes2;
	TParaAttributes2* FParaAttributes2;

#ifdef USERICHEDITOLE
	TIRichEditOle* FRichEditOle;
#endif

	virtual void __fastcall CreateParams(Controls::TCreateParams &Params);
	virtual void __fastcall DestroyWnd(void);
	AnsiString BuildUndoName(AnsiString typeName, int id);
	virtual void __fastcall CreateWnd(void);
	bool __fastcall GetCanUndo(void);
	bool __fastcall GetCanRedo(void);
	void __fastcall SetUndoLimit(int limit);
	AnsiString __fastcall GetUndoName(void);
	AnsiString __fastcall GetRedoName(void);
	void __fastcall SetMultiUndo(bool multiLevel);
	CHARRANGE __fastcall GetSelRange(void);
//	__property WordWrap;			// hide property (still visible in editor?)

	virtual void __fastcall WndProc(Messages::TMessage &Message);
	unsigned long int GetCurrentLength(void);
	void __fastcall SetInsertMode(bool insertMode);
	void ToggleInsertMode(void);
	void __fastcall SetAutoUrlDetect(bool autoDetect);

	void __fastcall SetSelAttributes2(TTextAttributes2* attr);
	void __fastcall SetDefAttributes2(TTextAttributes2* attr);

public:
	__fastcall TRichEdit20(TComponent* Owner);
	__fastcall ~TRichEdit20();
	bool Undo(void);
	bool Redo(void);
	void StopGroupTyping(void);

	void ScrollCaret(void);
	bool FindTextEx(AnsiString text, int startPos, TFindOptions findOptions,
		int& foundPos, int& length);
	void __fastcall GetExSel(int &SelStart, int &SelStop);
	void __fastcall SetExSel(int SelStart, int SelStop);
	void __fastcall ReplaceSelectedText(AnsiString text, bool canUndo = true);
#ifdef USERICHEDITOLE
	bool __fastcall InsertObject(void);
	bool __fastcall PasteSpecial(void);
	void __fastcall PasteSpecialFromClipboard(char* format = CF_RTF) {
		PasteSpecialFromClipboard((unsigned int) format);
		};
	void __fastcall PasteSpecialFromClipboard(unsigned int format);
#endif
	bool CanPasteFormat(char* format) { return CanPasteFormat((unsigned int) format); };
	bool CanPasteFormat(unsigned int format);

	virtual void __fastcall RecreateWnd(void);	// not virtual in base!!!
	void WorkAroundVerticalScrollBarBug(void);

	int GetTabStops(TTabStops& tabStops);
	void SetTabStops(int tabCount, TTabStops& tabStops);
	void SetTabStops(int tabSize);
	void SetTabStops(int tabChars, TFont& font, bool entireDocument = true);

	TRect TwipsToTargetRect(HDC hdc, TRect rect);

#ifdef USERICHEDITOLE
	void SetOleHostNames(AnsiString hostApp, AnsiString hostDoc);
#endif
	DYNAMIC void __fastcall KeyDown(Word &Key, Classes::TShiftState Shift);

	unsigned int GetLine(void);
	void SetLine(unsigned int line);
	unsigned int GetColumn(void);
	void SetColumn(unsigned int column);

#ifdef USERICHEDITOLE
	bool CloseActiveObjects(bool savePrompt);
#endif

	static bool IsRTF(AnsiString filepath);
	static bool IsLikelyBinaryFile(AnsiString filepath);

	__property bool SupportsMultiUndo = { read = FSupportsMultiUndo, nodefault };
	__property AnsiString UndoName = { read = GetUndoName };
	__property AnsiString RedoName = { read = GetRedoName };
	__property bool CanUndo = { read = GetCanUndo };
	__property bool CanRedo = { read = GetCanRedo };
	__property CHARRANGE SelRange = { read = GetSelRange };
	__property unsigned long int CurrentLength = { read = GetCurrentLength, nodefault };
	__property bool InsertMode = { read = FInsertMode, write = SetInsertMode, nodefault };
	__property unsigned int Line = { read = GetLine, write = SetLine, nodefault };
	__property unsigned int Column = { read = GetColumn, write = SetColumn, nodefault };
	__property TTextAttributes2* SelAttributes2 = { read = FSelAttributes2, write = SetSelAttributes2, default = 0};
	__property TTextAttributes2* DefAttributes2 = { read = FDefAttributes2, write = SetDefAttributes2, default = 0};
	__property TParaAttributes2* Paragraph2 = { read = FParaAttributes2, write = FParaAttributes2, default = 0 };

__published:
	__property int UndoLimit = { read = FUndoLimit, write = SetUndoLimit, default = 100 };
	__property bool MultiUndo = { read = FMultiUndo, write = SetMultiUndo, nodefault };
	__property TNotifyEvent OnInsertChange = { read = FOnInsertChange, write = FOnInsertChange, default = 0 };
	__property bool AutoUrlDetect = { read = FAutoUrlDetect, write = SetAutoUrlDetect, default = false };
};
//---------------------------------------------------------------------------
#endif
//===========================================================================
// Copyright  1998 Thin Air Enterprises and Robert Dunn
//===========================================================================
